# -*- coding: utf-8 -*-
#
# Copyright (C) 2023-2024 CERN.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Unit of work."""

from invenio_records_resources.services.uow import RecordCommitOp


class ParentRecordCommitOp(RecordCommitOp):
    """Parent record commit operation, bulk indexing records and drafts."""

    def __init__(
        self,
        parent,
        indexer_context=None,
        bulk_index=True,
    ):
        """Initialize the parent record commit operation.

        :params indexer_context: a dict containing record/draft cls and indexers,
            or service. Expected keys: `record_cls, draft_cls, indexer, draft_indexer`.
            `None` if indexing disabled.
        :params bulk_index: if ``True`` will send to bulk indexing queue. Otherwise will
            index synchronously.
        """
        super().__init__(parent, indexer=None)
        self._indexer_context = indexer_context
        if indexer_context:
            _service = indexer_context.get("service", None)
            if _service is not None:
                self._record_cls = getattr(_service, "record_cls")
                self._draft_cls = getattr(_service, "draft_cls")
                self._record_indexer = getattr(_service, "indexer")
                self._draft_indexer = getattr(_service, "draft_indexer")
            else:
                self._record_cls = indexer_context["record_cls"]
                self._draft_cls = indexer_context["draft_cls"]
                self._record_indexer = indexer_context["indexer"]
                self._draft_indexer = indexer_context["draft_indexer"]
        self._bulk_index = bulk_index

    def _get_siblings(self):
        """Get all record and draft versions by parent.

        This operation might be slow when a record has a high number of versions and drafts.
        However, given that often versions should be re-indexed as soon as possible when the parent
        is committed, the fetching operation is done synchronously.
        """
        records_ids = self._record_cls.get_records_by_parent(
            self._record, with_deleted=False, ids_only=True
        )
        drafts_ids = self._draft_cls.get_records_by_parent(
            self._record, with_deleted=False, ids_only=True
        )
        return records_ids, drafts_ids

    def on_commit(self, uow):
        """No commit operation."""
        pass

    def on_post_commit(self, uow):
        """Bulk index as last operation."""
        if self._indexer_context is not None:
            records_ids, drafts_ids = self._get_siblings()
            if records_ids:
                if self._bulk_index:
                    self._record_indexer.bulk_index(records_ids)
                else:
                    for record_id in records_ids:
                        self._record_indexer.index_by_id(record_id)
            if drafts_ids:
                if self._bulk_index:
                    self._draft_indexer.bulk_index(drafts_ids)
                else:
                    for draft_id in drafts_ids:
                        self._draft_indexer.index_by_id(draft_id)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Service tests.

Test to add:
- Read a tombstone page
- Read with missing permissions
- Read with missing pid
"""

from datetime import timedelta

import pytest


#
# Fixtures
#
@pytest.fixture()
def input_data(input_data):
    """Enable files."""
    input_data["files"]["enabled"] = False
    return input_data


def test_hard_delete_soft_deleted(app, service, identity_simple, input_data):
    draft = service.create(identity_simple, input_data)
    service.publish(identity_simple, draft.id)
    draft_model = service.draft_cls.model_cls

    assert (
        len(draft_model.query.filter(draft_model.is_deleted == True).all()) == 1  # noqa
    )
    service.cleanup_drafts(timedelta(seconds=0), search_gc_deletes=0)

    assert (
        len(draft_model.query.filter(draft_model.is_deleted == True).all()) == 0  # noqa
    )


def test_hard_delete_soft_deleted_not_enough_time(
    app, service, identity_simple, input_data
):
    draft = service.create(identity_simple, input_data)
    service.publish(identity_simple, draft.id)
    draft_model = service.draft_cls.model_cls

    assert (
        len(draft_model.query.filter(draft_model.is_deleted == True).all()) == 1  # noqa
    )
    service.cleanup_drafts(timedelta(seconds=10), search_gc_deletes=0)

    assert (
        len(draft_model.query.filter(draft_model.is_deleted == True).all()) == 1  # noqa
    )



"""Example of a record model."""

from invenio_db import db
from invenio_files_rest.models import Bucket
from invenio_records.models import RecordMetadataBase
from invenio_records_resources.records import FileRecordModelMixin
from sqlalchemy_utils.types import UUIDType

from invenio_drafts_resources.records import (
    DraftMetadataBase,
    ParentRecordMixin,
    ParentRecordStateMixin,
)


class ParentRecordMetadata(db.Model, RecordMetadataBase):
    """Model for mock module metadata."""

    __tablename__ = "parent_mock_metadata"


class DraftMetadata(db.Model, DraftMetadataBase, ParentRecordMixin):
    """Model for mock module metadata."""

    __tablename__ = "draft_mock_metadata"
    __parent_record_model__ = ParentRecordMetadata

    bucket_id = db.Column(UUIDType, db.ForeignKey(Bucket.id), index=True)
    bucket = db.relationship(Bucket, foreign_keys=[bucket_id])


class RecordMetadata(db.Model, RecordMetadataBase, ParentRecordMixin):
    """Model for mock module metadata."""

    __tablename__ = "record_mock_metadata"
    __parent_record_model__ = ParentRecordMetadata

    bucket_id = db.Column(UUIDType, db.ForeignKey(Bucket.id), index=True)
    bucket = db.relationship(Bucket, foreign_keys=[bucket_id])


class ParentState(db.Model, ParentRecordStateMixin):
    """Model for mock module for parent state."""

    __parent_record_model__ = ParentRecordMetadata
    __record_model__ = RecordMetadata
    __draft_model__ = DraftMetadata


class FileRecordMetadata(db.Model, RecordMetadataBase, FileRecordModelMixin):
    """Model for mock module record files."""

    __record_model_cls__ = RecordMetadata

    __tablename__ = "mock_record_files"


class MediaFileRecordMetadata(db.Model, RecordMetadataBase, FileRecordModelMixin):
    """Model for mock module record files."""

    __record_model_cls__ = RecordMetadata

    __tablename__ = "mock_record_media_files"


class FileDraftMetadata(db.Model, RecordMetadataBase, FileRecordModelMixin):
    """Model for mock module draft files."""

    __record_model_cls__ = DraftMetadata

    __tablename__ = "mock_draft_files"


class MediaFileDraftMetadata(db.Model, RecordMetadataBase, FileRecordModelMixin):
    """File associated with a draft."""

    __record_model_cls__ = DraftMetadata

    __tablename__ = "mock_drafts_media_files"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Records service component base classes."""
from invenio_i18n import gettext as _
from invenio_records_resources.services.base.config import _make_cls
from marshmallow import ValidationError

from .base import BaseRecordFilesComponent


class _DraftMediaFilesComponent(BaseRecordFilesComponent):
    """Draft media files component."""

    def create(self, identity, data=None, record=None, errors=None):
        """Assigns files.enabled.

        NOTE: `record` actually refers to the draft
              (this interface is used in records-resources and rdm-records)
        """
        draft = record
        files = self.get_record_files(draft)

        # disable the media files by default
        enabled = data.get(self.files_data_key, {}).get(
            "enabled", self.service.config.default_media_files_enabled
        )

        files.enabled = enabled

    def update_draft(self, identity, data=None, record=None, errors=None):
        """Assigns files.enabled and warns if files are missing.

        NOTE: `record` actually refers to the draft
              (this interface is used in records-resources and rdm-records)
        """
        draft = record
        draft_files = self.get_record_files(draft)
        record_files = self.get_record_files(record)
        enabled = data.get(self.files_data_key, {}).get("enabled", record_files.enabled)
        default_preview = data.get(self.files_data_key, {}).get("default_preview")

        try:
            self.assign_files_enabled(draft, enabled)
        except ValidationError as e:
            errors.append(
                {"field": f"{self.files_data_key}.enabled", "messages": e.messages}
            )
            return  # exit early

        if draft_files.enabled and not draft_files.items():
            errors.append(
                {
                    "field": f"{self.files_data_key}.enabled",
                    "messages": [
                        _(
                            "Missing uploaded files. To disable files for "
                            "this record please mark it as metadata-only."
                        )
                    ],
                }
            )

        try:
            self.assign_files_default_preview(
                draft,
                default_preview,
            )
        except ValidationError as e:
            errors.append(
                {
                    "field": f"{self.files_data_key}.default_preview",
                    "messages": e.messages,
                }
            )

    def _publish_new(self, identity, draft, record):
        """Action when publishing a new draft."""
        # For unpublished drafts (new and new version), we move the draft
        # bucket from the draft to the record (instead of creating a new, and
        # deleting one). For consistency, we keep a bucket for all records
        # independently of if they have files enabled or not.
        draft_files = self.get_record_files(draft)
        draft_bucket = self.get_record_bucket(draft)

        if draft_bucket is None:
            # legacy compatibility safety check - create bucket if missing
            draft_files.create_bucket()
            draft_bucket = self.get_record_bucket(draft)

        super()._publish_new(identity, draft, record)

    def edit(self, identity, draft=None, record=None):
        """Edit callback."""
        draft_files = self.get_record_files(draft)
        record_files = self.get_record_files(record)
        draft_bucket = self.get_record_bucket(draft)
        if draft_bucket is None:
            # Happens, when a soft-deleted draft is un-deleted.
            draft[self.files_data_key] = {"enabled": record_files.enabled}
            # re-fetch the files attribute - above data getter sets the attribute
            draft_files = self.get_record_files(draft)
            draft_files.create_bucket()
        # we copy always file objects and tear them down when publish
        draft_files.copy(record_files)
        # in the media files we don't lock the bucket
        # - they can be simply edited from draft

    def _publish_edit(self, identity, draft, record):
        """Action when publishing an edit to an existing record."""
        record_files = self.get_record_files(record)
        draft_files = self.get_record_files(draft)

        record_files.unlock()
        # we make sure that record files are enabled according to draft because
        # media files can be enabled after the first publish and thus the record will
        # have an outdated value
        record_files.enabled = draft_files.enabled
        record_files.sync(draft_files, delete_extras=True)
        record_files.lock()
        # Teardown the bucket and files created in edit().
        self._purge_bucket_and_ovs(draft_files)

    def new_version(self, identity, draft=None, record=None):
        """New version callback."""
        # We don't copy files from the previous version, but instead allow
        # users to import the files.
        draft_files = self.get_record_files(draft)
        # Set media files to the default config
        draft_files.enabled = self.service.config.default_media_files_enabled

    def import_files(self, identity, draft=None, record=None):
        """Import files callback."""
        # We don't need the import for media files
        return


MediaFilesAttrConfig = {
    "_files_attr_key": "media_files",
    "_files_data_key": "media_files",
    "_files_bucket_attr_key": "media_bucket",
    "_files_bucket_id_attr_key": "media_bucket_id",
}

### Configure file attributes for media files component
DraftMediaFilesComponent = _make_cls(
    _DraftMediaFilesComponent, {**MediaFilesAttrConfig}
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2023 CERN.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Versions field."""

import uuid

from invenio_db import db
from invenio_records.systemfields import SystemFieldContext
from invenio_records.systemfields.base import SystemField


def uuid_or_none(val):
    """Convert string to UUID object."""
    if val is not None and not isinstance(val, uuid.UUID):
        return uuid.UUID(val)
    return val


class VersionsManager:
    """Versions state manager."""

    def __init__(self, record, dump=None):
        """Initialize the versions manager."""
        self._record = record
        self._state = None
        if dump is not None:
            self.load(dump)

    def copy_to(self, record):
        """Create a copy of the version manager and set on another record."""
        record.model.index = self.index
        version_manager = self.__class__(record)
        version_manager._state = self._state
        return version_manager

    #
    # Record managed attributes
    #
    @property
    def model_cls(self):
        """Get versions state management model."""
        return self._record.versions_model_cls

    @property
    def parent_id(self):
        """Get versions state management model."""
        return self._record.model.parent_id

    @property
    def record_model_cls(self):
        """Get model cls of the record/draft."""
        return self._record.model_cls

    @property
    def index(self):
        """Get the version index of the record/draft."""
        return self._record.model.index

    #
    # Parent managed attributes
    #
    @property
    def latest_id(self):
        """The id of the latest published record/draft."""
        return self.state().latest_id

    @property
    def latest_index(self):
        """The version index of the latest published record/draft."""
        return self.state().latest_index

    @property
    def next_draft_id(self):
        """The id of the next draft (and record)."""
        return self.state().next_draft_id

    #
    # Computed attributes
    #
    @property
    def is_latest(self):
        """Check if the record/draft id is the latest published record id."""
        return self.latest_id == self._record.id

    @property
    def is_latest_draft(self):
        """Check if the record/draft id is the latest draft id."""
        if self.next_draft_id:
            return self.next_draft_id == self._record.id
        else:
            return self.latest_id == self._record.id

    @property
    def next_index(self):
        """Get the next parent index."""
        latest_index_by_parent = None
        with db.session.no_autoflush:
            rec_model = (
                self.record_model_cls.query.filter_by(parent_id=self.parent_id)
                .order_by(self.record_model_cls.index.desc())
                .first()
            )
            if rec_model:
                latest_index_by_parent = rec_model.index
        return latest_index_by_parent + 1 if latest_index_by_parent is not None else 1

    #
    # State management methods
    #
    def state(self, refresh=False):
        """Retrieve the versions state."""
        if self._state is None or refresh:
            # Get object if it exists
            self._state = self.model_cls.query.filter_by(
                parent_id=self.parent_id
            ).one_or_none()
            if self._state is None:
                # Object doesn't exists, so create it.
                self._state = self.model_cls(parent_id=self.parent_id)
                db.session.add(self._state)
        return self._state

    def set_next(self):
        """Set this record as the next draft."""
        self.state().next_draft_id = self._record.id
        self._record.model.index = self.next_index

    def clear_next(self):
        """Unset this record as the next draft."""
        self.state().next_draft_id = None
        self._record.model.index = None

    def set_latest(self):
        """Set this record as the latest published record."""
        self.state().latest_id = self._record.id
        self.state().latest_index = self.index
        self.state().next_draft_id = None

    #
    # Dump/load
    #
    def dump(self):
        """Dump the versions state to the index."""
        return dict(
            latest_id=self.latest_id,
            latest_index=self.latest_index,
            next_draft_id=self.next_draft_id,
            is_latest=self.is_latest,
            is_latest_draft=self.is_latest_draft,
            index=self.index,
        )

    def load(self, dump):
        """Load the state."""
        self._state = self.model_cls(
            parent_id=self.parent_id,
            latest_id=uuid_or_none(dump["latest_id"]),
            latest_index=dump["latest_index"],
            next_draft_id=uuid_or_none(dump["next_draft_id"]),
        )
        if self.index != dump["index"]:
            self._record.model.index = dump["index"]

    def __repr__(self):
        """Return repr(self)."""
        return (
            f"<{type(self).__name__} (parent_id: {self.parent_id}, "
            f"index: {self.index}, latest_id: {self.latest_id}, "
            f"latest_index: {self.latest_index}, "
            f"next_draft_id: {self.next_draft_id})>"
        )


class VersionsFieldContext(SystemFieldContext):
    """Version field context."""

    @property
    def model_cls(self):
        """Versions model class."""
        return self.record_cls.versions_model_cls

    def resolve(self, *, parent_id):
        """Resolve a versions state from a parent ID."""
        return self.model_cls.query.filter_by(parent_id=parent_id).one()


class VersionsField(SystemField):
    """Versions field."""

    def __init__(self, create=True, set_next=False, set_latest=False):
        """Initialise the versions field."""
        self._create = create
        self._set_next = set_next
        self._set_latest = set_latest
        super().__init__()

    def obj(self, record):
        """Get the version manager."""
        obj = self._get_cache(record)
        if obj is not None:
            return obj
        obj = VersionsManager(record)
        self._set_cache(record, obj)
        return obj

    def set_obj(self, record, versions):
        """Set an version manager on the record."""
        assert isinstance(versions, VersionsManager)
        versions = versions.copy_to(record)
        self._set_cache(record, versions)

    #
    # Record life-cycle hooks
    #
    def post_create(self, record):
        """Called after a record is created."""
        # The parent record is created on pre_create.
        versions = self.obj(record)
        if self._create and self._set_next:
            if not record.is_published:
                versions.set_next()
        elif self._create and self._set_latest:
            versions.set_latest()

    def pre_delete(self, record, force=False):
        """Called before a record is deleted."""
        if force:
            versions = self.obj(record)
            versions.clear_next()

    def pre_dump(self, record, data, dumper=None):
        """Called before a record is dumped."""
        parent = getattr(record, self.attr_name)
        if parent:
            data[self.attr_name] = self.obj(record).dump()

    def post_load(self, record, data, loader=None):
        """Called after a record was loaded."""
        dump = record.pop(self.attr_name, None)
        if dump:
            versions = VersionsManager(record, dump=dump)
            setattr(record, self.attr_name, versions)

    #
    # Data descriptor API
    #
    def __get__(self, record, owner=None):
        """Get the value for the field.

        Called when the field is accessed, e.g:

        .. code-block:: python

            # Access by object
            record.versions
            # Access by object
            Record.versions
        """
        if record is None:
            # access by class
            return VersionsFieldContext(self, owner)
        # access by object
        return self.obj(record)

    def __set__(self, record, obj):
        """Assign a value to the field.

        Called when a value is assigned to the field, e.g.:

        .. code-block:: python

            record.versions = <obj>
        """
        self.set_obj(record, obj)
